--- llseek.c	Sun Jun 08 02:10:42 1997
+++ new/llseek.c	Wed Aug 06 23:18:46 1997
@@ -1,18 +1,37 @@
 /*
  * llseek.c -- stub calling the llseek system call
  *
- * Copyright (C) 1994 Remy Card.  This file may be redistributed
- * under the terms of the GNU Public License.
+ * Copyright (C) 1994, 1995, 1996 Theodore Ts'o.
+ *
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Public
+ * License.
+ * %End-Header%
  */
 
+#include <sys/types.h>
+
 #include <errno.h>
-#include <string.h>
-#include <linux/types.h>
+#include <unistd.h>
 #include "et/com_err.h"
-#include "io.h"
+#include "ext2fs/io.h"
 
 #ifdef __linux__
 
+#ifdef HAVE_LLSEEK
+#include <unistd.h>
+#include <syscall.h>
+
+#else	/* HAVE_LLSEEK */
+
+#ifdef __alpha__
+
+#define llseek lseek
+
+#else /* !__alpha__ */
+
+#include <linux/unistd.h>
+
 #ifndef __NR__llseek
 #define __NR__llseek            140
 #endif
@@ -24,62 +43,62 @@
 		 unsigned long, offset_low,ext2_loff_t *,result,
 		 unsigned int, origin)
 
-ext2_loff_t ext2_llseek (unsigned int fd, ext2_loff_t offset,
+static ext2_loff_t llseek (unsigned int fd, ext2_loff_t offset,
 			 unsigned int origin)
 {
-	unsigned long offset_high;
-	unsigned long offset_low;
 	ext2_loff_t result;
 	int retval;
+
+	retval = _llseek (fd, ((unsigned long long) offset) >> 32,
+			((unsigned long long) offset) & 0xffffffff,
+			&result, origin);
+	return (retval == -1 ? (ext2_loff_t) retval : result);
+}
+
+#endif	/* HAVE_LLSEEK */
+
+#endif /* __alpha__ */
+
+ext2_loff_t ext2fs_llseek (unsigned int fd, ext2_loff_t offset,
+			 unsigned int origin)
+{
+	ext2_loff_t result;
 	static int do_compat = 0;
 
-	if (do_compat)
-		return lseek (fd, (off_t) offset, origin);
+	if ((sizeof(off_t) >= sizeof(ext2_loff_t)) ||
+	    (offset < ((ext2_loff_t) 1 << ((sizeof(off_t)*8) -1))))
+		return lseek(fd, (off_t) offset, origin);
 	
-	offset_high = ((unsigned long long) offset) >> 32;
-	offset_low = ((unsigned long long) offset) & 0xffffffff;
-	retval = _llseek (fd, offset_high, offset_low, &result, origin);
-	if (retval == -1 && errno == ENOSYS) {
+	if (do_compat) {
+		errno = EINVAL;
+		return -1;
+	}
+	
+	result = llseek (fd, offset, origin);
+	if (result == -1 && errno == ENOSYS) {
 		/*
 		 * Just in case this code runs on top of an old kernel
 		 * which does not support the llseek system call
 		 */
 		do_compat++;
-		return lseek (fd, (off_t) offset, origin);
+		errno = EINVAL;
 	}
-	if (retval == -1)
-		result = -1;
 	return result;
 }
 
-#else
-
-#if 0
-#include <windows.h>
-#include "winperror.h"
-#endif
+#else /* !linux */
 
-ext2_loff_t ext2_llseek (unsigned int fd, ext2_loff_t offset,
+ext2_loff_t ext2fs_llseek (unsigned int fd, ext2_loff_t offset,
 			 unsigned int origin)
 {
-#if 0
-/*
 	if ((sizeof(off_t) < sizeof(ext2_loff_t)) &&
 	    (offset >= ((ext2_loff_t) 1 << ((sizeof(off_t)*8) -1)))) {
-		errno = -EINVAL;
+		errno = EINVAL;
 		return -1;
 	}
 	return lseek (fd, (off_t) offset, origin);
-*/
-	DWORD dwPointer;
-
-	dwPointer = SetFilePointer ((HANDLE)fd, offset, NULL, FILE_BEGIN);
-	if ( dwPointer == 0xffffffff ) {
-		DWORD err = winperror("SetFilePointer error");
-        return err;
-#endif
-	errno = -EINVAL;
-	return -1;
 }
 
-#endif
+#endif 	/* linux */
+
+
